@@ -9,22 +9,20 @@ header files (so that the JS compiler can see the constants in those
 headers, for the libc implementation in JS).
 """
 
-from __future__ import print_function
-
 import os
 import json
 import subprocess
-import shutil
 import time
 import logging
 import pprint
+import shutil
 from collections import OrderedDict
 
 from tools import building
 from tools import diagnostics
 from tools import shared
 from tools import gen_struct_info
-from tools.response_file import substitute_response_files
+from tools import webassembly
 from tools.shared import WINDOWS, asstr, path_from_root, exit_with_error, asmjs_mangle, treat_as_user_function
 from tools.toolchain_profiler import ToolchainProfiler
 
@@ -47,7 +45,7 @@ def compute_minimal_runtime_initializer_and_exports(post, initializers, exports,
   exports_that_are_not_initializers = [x for x in exports if x not in initializers]
   # In Wasm backend the exports are still unmangled at this point, so mangle the names here
   exports_that_are_not_initializers = [asmjs_mangle(x) for x in exports_that_are_not_initializers]
-  post = post.replace('/*** ASM_MODULE_EXPORTS_DECLARES ***/', 'var ' + ','.join(exports_that_are_not_initializers) + ';')
+  post = post.replace('/*** ASM_MODULE_EXPORTS_DECLARES ***/', 'var ' + ',\n  '.join(exports_that_are_not_initializers) + ';')
 
   # Generate assignments from all asm.js/wasm exports out to the JS variables above: e.g. a = asm['a']; b = asm['b'];
   post = post.replace('/*** ASM_MODULE_EXPORTS ***/', receiving)
@@ -80,18 +78,16 @@ def optimize_syscalls(declares, DEBUG):
     # without filesystem support, it doesn't matter what syscalls need
     shared.Settings.SYSCALLS_REQUIRE_FILESYSTEM = 0
   else:
-    syscall_prefixes = ('__sys', 'fd_', '__wasi_fd_')
+    syscall_prefixes = ('__sys', 'fd_')
     syscalls = [d for d in declares if d.startswith(syscall_prefixes)]
     # check if the only filesystem syscalls are in: close, ioctl, llseek, write
     # (without open, etc.. nothing substantial can be done, so we can disable
     # extra filesystem support in that case)
     if set(syscalls).issubset(set([
       '__sys_ioctl',
-      # legacy/fastcomp name for __sys_ioctl
-      '__syscall6',
-      'fd_seek', '__wasi_fd_seek',
-      'fd_write', '__wasi_fd_write',
-      'fd_close', '__wasi_fd_close',
+      'fd_seek',
+      'fd_write',
+      'fd_close',
     ])):
       if DEBUG:
         logger.debug('very limited syscalls (%s) so disabling full filesystem support', ', '.join(str(s) for s in syscalls))
@@ -145,17 +141,25 @@ def update_settings_glue(metadata, DEBUG):
 
     shared.Settings.PROXIED_FUNCTION_SIGNATURES = read_proxied_function_signatures(metadata['asmConsts'])
 
-  metadata['staticBump'] = align_memory(metadata['staticBump'])
-
   shared.Settings.BINARYEN_FEATURES = metadata['features']
-  shared.Settings.WASM_TABLE_SIZE = metadata['tableSize']
   if shared.Settings.RELOCATABLE:
     # When building relocatable output (e.g. MAIN_MODULE) the reported table
     # size does not include the reserved slot at zero for the null pointer.
     # Instead we use __table_base to offset the elements by 1.
-    shared.Settings.WASM_TABLE_SIZE += 1
+    shared.Settings.WASM_TABLE_SIZE = metadata['tableSize'] + 1
   shared.Settings.MAIN_READS_PARAMS = metadata['mainReadsParams']
 
+  # Store exports for Closure compiler to be able to track these as globals in
+  # -s DECLARE_ASM_MODULE_EXPORTS=0 builds.
+  shared.Settings.MODULE_EXPORTS = [(asmjs_mangle(f), f) for f in metadata['exports']]
+
+  if shared.Settings.STACK_OVERFLOW_CHECK:
+    if 'emscripten_stack_get_end' not in metadata['exports']:
+      logger.warning('STACK_OVERFLOW_CHECK disabled because emscripten stack helpers not exported')
+      shared.Settings.STACK_OVERFLOW_CHECK = 0
+    else:
+      shared.Settings.EXPORTED_RUNTIME_METHODS += ['writeStackCookie', 'checkStackCookie']
+
 
 # static code hooks
 class StaticCodeHooks:
@@ -201,43 +205,20 @@ def compile_settings(temp_files):
 
 
 class Memory():
-  def __init__(self, metadata):
-    # Note: if RELOCATABLE, then only relative sizes can be computed, and we don't
-    #       actually write out any absolute memory locations ({{{ STACK_BASE }}}
-    #       does not exist, etc.)
-
-    # Memory layout:
-    #  * first the static globals
-    self.global_base = shared.Settings.GLOBAL_BASE
-    self.static_bump = metadata['staticBump']
-    #  * then the stack (up on fastcomp, down on upstream)
-    self.stack_low = align_memory(self.global_base + self.static_bump)
-    self.stack_high = align_memory(self.stack_low + shared.Settings.TOTAL_STACK)
-    self.stack_base = self.stack_high
-    self.stack_max = self.stack_low
-    #  * then dynamic memory begins
-    self.dynamic_base = align_memory(self.stack_high)
+  def __init__(self, static_bump):
+    stack_low = align_memory(shared.Settings.GLOBAL_BASE + static_bump)
+    stack_high = align_memory(stack_low + shared.Settings.TOTAL_STACK)
+    self.stack_base = stack_high
+    self.stack_max = stack_low
+    self.dynamic_base = align_memory(stack_high)
 
 
-def apply_memory(js, metadata):
+def apply_memory(js, memory):
   # Apply the statically-at-compile-time computed memory locations.
-  memory = Memory(metadata)
-
   # Write it all out
+  js = js.replace('{{{ HEAP_BASE }}}', str(memory.dynamic_base))
   js = js.replace('{{{ STACK_BASE }}}', str(memory.stack_base))
   js = js.replace('{{{ STACK_MAX }}}', str(memory.stack_max))
-  js = js.replace('{{{ DYNAMIC_BASE }}}', str(memory.dynamic_base))
-
-  logger.debug('global_base: %d stack_base: %d, stack_max: %d, dynamic_base: %d, static bump: %d', memory.global_base, memory.stack_base, memory.stack_max, memory.dynamic_base, memory.static_bump)
-
-  shared.Settings.DYNAMIC_BASE = memory.dynamic_base
-
-  return js
-
-
-def apply_table(js):
-  js = js.replace('{{{ WASM_TABLE_SIZE }}}', str(shared.Settings.WASM_TABLE_SIZE))
-
   return js
 
 
@@ -247,9 +228,9 @@ def report_missing_symbols(all_implemented, pre):
   missing = list(set(shared.Settings.USER_EXPORTED_FUNCTIONS) - all_implemented)
 
   for requested in missing:
-    if ('function ' + asstr(requested)) in pre:
+    if ('function ' + asstr(requested) + '(') in pre:
       continue
-    diagnostics.warning('undefined', 'undefined exported function: "%s"', requested)
+    diagnostics.warning('undefined', 'undefined exported symbol: "%s"', requested)
 
   # Special hanlding for the `_main` symbol
 
@@ -308,104 +289,46 @@ def trim_asm_const_body(body):
   return body
 
 
-def create_fp_accessors(metadata):
-  if not shared.Settings.RELOCATABLE:
-    return ''
-
-  # Create `fp$XXX` handlers for determining function pionters (table addresses)
-  # at runtime.
-  # For SIDE_MODULEs these are generated by the proxyHandler at runtime.
-  accessors = []
-  for fullname in metadata['declares']:
-    if not fullname.startswith('fp$'):
-      continue
-    _, name, sig = fullname.split('$')
-    mangled = asmjs_mangle(name)
-    side = 'parent' if shared.Settings.SIDE_MODULE else ''
-    assertion = ('\n  assert(%sModule["%s"] || typeof %s !== "undefined", "external function `%s` is missing.' % (side, mangled, mangled, name) +
-                 'perhaps a side module was not linked in? if this symbol was expected to arrive '
-                 'from a system library, try to build the MAIN_MODULE with '
-                 'EMCC_FORCE_STDLIBS=XX in the environment");')
-    # the name of the original function is generally the normal function
-    # name, unless it is legalized, in which case the export is the legalized
-    # version, and the original provided by orig$X
-    if shared.Settings.LEGALIZE_JS_FFI and not shared.JS.is_legal_sig(sig):
-      name = 'orig$' + name
-
-    accessors.append('''
-Module['%(full)s'] = function() {
-  %(assert)s
-  // Use the original wasm function itself, for the table, from the main module.
-  var func = Module['asm']['%(original)s'];
-  // Try an original version from a side module.
-  if (!func) func = Module['_%(original)s'];
-  // Otherwise, look for a regular function or JS library function.
-  if (!func) func = Module['%(mangled)s'];
-  if (!func) func = %(mangled)s;
-  var fp = addFunction(func, '%(sig)s');
-  Module['%(full)s'] = function() { return fp };
-  return fp;
-}
-''' % {'full': asmjs_mangle(fullname), 'mangled': mangled, 'original': name, 'assert': assertion, 'sig': sig})
-
-  return '\n'.join(accessors)
-
-
 def create_named_globals(metadata):
-  if not shared.Settings.RELOCATABLE:
-    named_globals = []
-    for k, v in metadata['namedGlobals'].items():
-      # We keep __data_end alive internally so that wasm-emscripten-finalize knows where the
-      # static data region ends.  Don't export this to JS like other user-exported global
-      # address.
-      if k not in ['__data_end']:
-        named_globals.append("Module['_%s'] = %s;" % (k, v))
-    return '\n'.join(named_globals)
+  named_globals = []
+  for k, v in metadata['namedGlobals'].items():
+    v = int(v)
+    if shared.Settings.RELOCATABLE:
+      v += shared.Settings.GLOBAL_BASE
+    mangled = asmjs_mangle(k)
+    if shared.Settings.MINIMAL_RUNTIME:
+      named_globals.append("var %s = %s;" % (mangled, v))
+    else:
+      named_globals.append("var %s = Module['%s'] = %s;" % (mangled, mangled, v))
 
-  named_globals = '''
-var NAMED_GLOBALS = {
-  %s
-};
-for (var named in NAMED_GLOBALS) {
-  Module['_' + named] = gb + NAMED_GLOBALS[named];
-}
-Module['NAMED_GLOBALS'] = NAMED_GLOBALS;
-''' % ',\n  '.join('"' + k + '": ' + str(v) for k, v in metadata['namedGlobals'].items())
-
-  # wasm side modules are pure wasm, and cannot create their g$..() methods, so we help them out
-  # TODO: this works if we are the main module, but if the supplying module is later, it won't, so
-  #       we'll need another solution for that. one option is to scan the module imports, if/when
-  #       wasm supports that, then the loader can do this.
-  named_globals += '''
-for (var named in NAMED_GLOBALS) {
-  (function(named) {
-    var addr = Module['_' + named];
-    Module['g$_' + named] = function() { return addr };
-  })(named);
-}
-'''
-  named_globals += ''.join(["Module['%s'] = Module['%s'];\n" % (k, v) for k, v in metadata['aliases'].items()])
-  return named_globals
+  return '\n'.join(named_globals)
 
 
-def emscript(infile, outfile, memfile, temp_files, DEBUG):
+def emscript(in_wasm, out_wasm, outfile_js, memfile, temp_files, DEBUG):
   # Overview:
   #   * Run wasm-emscripten-finalize to extract metadata and modify the binary
   #     to use emscripten's wasm<->JS ABI
   #   * Use the metadata to generate the JS glue that goes with the wasm
 
-  metadata = finalize_wasm(temp_files, infile, outfile, memfile, DEBUG)
+  if shared.Settings.SINGLE_FILE:
+    # placeholder strings for JS glue, to be replaced with subresource locations in do_binaryen
+    shared.Settings.WASM_BINARY_FILE = '<<< WASM_BINARY_FILE >>>'
+  else:
+    # set file locations, so that JS glue can find what it needs
+    shared.Settings.WASM_BINARY_FILE = shared.JS.escape_for_js_string(os.path.basename(out_wasm))
+
+  metadata = finalize_wasm(in_wasm, out_wasm, memfile, DEBUG)
 
   update_settings_glue(metadata, DEBUG)
 
   if shared.Settings.SIDE_MODULE:
+    logger.debug('emscript: skipping remaining js glue generation')
     return
 
   if DEBUG:
     logger.debug('emscript: js compiler glue')
-
-  if DEBUG:
     t = time.time()
+
   glue, forwarded_data = compile_settings(temp_files)
   if DEBUG:
     logger.debug('  emscript: glue took %s seconds' % (time.time() - t))
@@ -419,43 +342,46 @@ def emscript(infile, outfile, memfile, temp_files, DEBUG):
 
   pre, post = glue.split('// EMSCRIPTEN_END_FUNCS')
 
-  # memory and global initializers
-
-  global_initializers = ', '.join('{ func: function() { %s() } }' % i for i in metadata['initializers'])
-
-  if shared.Settings.MINIMAL_RUNTIME:
-    # In minimal runtime, global initializers are run after the Wasm Module instantiation has finished.
-    global_initializers = ''
-  else:
-    # In regular runtime, global initializers are recorded in an __ATINIT__ array.
-    global_initializers = '''/* global initializers */ %s __ATINIT__.push(%s);
-''' % ('if (!ENVIRONMENT_IS_PTHREAD)' if shared.Settings.USE_PTHREADS else '',
-       global_initializers)
-
-  pre += '\n' + global_initializers + '\n'
-
-  pre = apply_memory(pre, metadata)
-  pre = apply_static_code_hooks(pre) # In regular runtime, atinits etc. exist in the preamble part
-  post = apply_static_code_hooks(post) # In MINIMAL_RUNTIME, atinit exists in the postamble part
-
-  if shared.Settings.RELOCATABLE and not shared.Settings.SIDE_MODULE:
-    pre += 'var gb = GLOBAL_BASE, fb = 0;\n'
-
-  # merge forwarded data
-  shared.Settings.EXPORTED_FUNCTIONS = forwarded_json['EXPORTED_FUNCTIONS']
-
   exports = metadata['exports']
 
-  # Store exports for Closure compiler to be able to track these as globals in
-  # -s DECLARE_ASM_MODULE_EXPORTS=0 builds.
-  shared.Settings.MODULE_EXPORTS = [(asmjs_mangle(f), f) for f in exports]
-
   if shared.Settings.ASYNCIFY:
     exports += ['asyncify_start_unwind', 'asyncify_stop_unwind', 'asyncify_start_rewind', 'asyncify_stop_rewind']
 
-  report_missing_symbols(set([asmjs_mangle(f) for f in exports]), pre)
+  all_exports = exports + list(metadata['namedGlobals'].keys())
+  all_exports = set([asmjs_mangle(e) for e in all_exports])
+  report_missing_symbols(all_exports, pre)
 
-  asm_consts = create_asm_consts_wasm(forwarded_json, metadata)
+  if not outfile_js:
+    logger.debug('emscript: skipping remaining js glue generation')
+    return
+
+  # memory and global initializers
+
+  # In minimal runtime, global initializers are run after the Wasm Module instantiation has finished.
+  if not shared.Settings.MINIMAL_RUNTIME:
+    global_initializers = ', '.join('{ func: function() { %s() } }' % i for i in metadata['initializers'])
+    # In regular runtime, global initializers are recorded in an __ATINIT__ array.
+    global_initializers = '__ATINIT__.push(%s);' % global_initializers
+    if shared.Settings.USE_PTHREADS:
+      global_initializers = 'if (!ENVIRONMENT_IS_PTHREAD) ' + global_initializers
+
+    pre += '\n' + global_initializers + '\n'
+
+  if shared.Settings.RELOCATABLE:
+    static_bump = align_memory(webassembly.parse_dylink_section(in_wasm)[0])
+    memory = Memory(static_bump)
+    logger.debug('stack_base: %d, stack_max: %d, dynamic_base: %d', memory.stack_base, memory.stack_max, memory.dynamic_base)
+
+    pre = apply_memory(pre, memory)
+    post = apply_memory(post, memory)
+
+  pre = apply_static_code_hooks(pre) # In regular runtime, atinits etc. exist in the preamble part
+  post = apply_static_code_hooks(post) # In MINIMAL_RUNTIME, atinit exists in the postamble part
+
+  # merge forwarded data
+  shared.Settings.EXPORTED_FUNCTIONS = forwarded_json['EXPORTED_FUNCTIONS']
+
+  asm_consts = create_asm_consts(metadata)
   em_js_funcs = create_em_js(forwarded_json, metadata)
   asm_const_pairs = ['%s: %s' % (key, value) for key, value in asm_consts]
   asm_const_map = 'var ASM_CONSTS = {\n  ' + ',  \n '.join(asm_const_pairs) + '\n};\n'
@@ -463,29 +389,23 @@ def emscript(infile, outfile, memfile, temp_files, DEBUG):
     '// === Body ===',
     ('// === Body ===\n\n' + asm_const_map +
      '\n'.join(em_js_funcs) + '\n'))
-  pre = apply_table(pre)
-  outfile.write(pre)
-  pre = None
 
-  invoke_funcs = metadata['invokeFuncs']
-  try:
-    del forwarded_json['Variables']['globals']['_llvm_global_ctors'] # not a true variable
-  except KeyError:
-    pass
+  with open(outfile_js, 'w') as out:
+    out.write(pre)
+    pre = None
 
-  sending = create_sending_wasm(invoke_funcs, forwarded_json, metadata)
-  receiving = create_receiving_wasm(exports, metadata['initializers'])
+    invoke_funcs = metadata['invokeFuncs']
+    sending = create_sending(invoke_funcs, metadata)
+    receiving = create_receiving(exports, metadata['initializers'])
 
-  if shared.Settings.MINIMAL_RUNTIME:
-    post = compute_minimal_runtime_initializer_and_exports(post, metadata['initializers'], exports, receiving)
-    receiving = ''
+    if shared.Settings.MINIMAL_RUNTIME:
+      post = compute_minimal_runtime_initializer_and_exports(post, metadata['initializers'], exports, receiving)
+      receiving = ''
 
-  module = create_module_wasm(sending, receiving, invoke_funcs, metadata)
+    module = create_module(sending, receiving, invoke_funcs, metadata)
 
-  write_output_file(outfile, post, module)
-  module = None
-
-  outfile.close()
+    write_output_file(out, post, module)
+    module = None
 
 
 def remove_trailing_zeros(memfile):
@@ -498,34 +418,25 @@ def remove_trailing_zeros(memfile):
     f.write(mem_data[:end])
 
 
-def finalize_wasm(temp_files, infile, outfile, memfile, DEBUG):
-  basename = shared.unsuffixed(outfile.name)
-  wasm = basename + '.wasm'
-  base_wasm = infile
+def finalize_wasm(infile, outfile, memfile, DEBUG):
   building.save_intermediate(infile, 'base.wasm')
-
   args = ['--detect-features', '--minimize-wasm-changes']
 
   # if we don't need to modify the wasm, don't tell finalize to emit a wasm file
   modify_wasm = False
 
-  # C++ exceptions and longjmp require invoke processing,
-  # https://github.com/WebAssembly/binaryen/issues/3081
-  if shared.Settings.SUPPORT_LONGJMP or shared.Settings.DISABLE_EXCEPTION_CATCHING != 1:
-    modify_wasm = True
   if shared.Settings.WASM2JS:
     # wasm2js requires full legalization (and will do extra wasm binary
     # later processing later anyhow)
     modify_wasm = True
-  write_source_map = shared.Settings.DEBUG_LEVEL >= 4
-  if write_source_map:
-    building.emit_wasm_source_map(base_wasm, base_wasm + '.map')
-    building.save_intermediate(base_wasm + '.map', 'base_wasm.map')
-    args += ['--output-source-map-url=' + shared.Settings.SOURCE_MAP_BASE + os.path.basename(shared.Settings.WASM_BINARY_FILE) + '.map']
+  if shared.Settings.GENERATE_SOURCE_MAP:
+    building.emit_wasm_source_map(infile, infile + '.map', outfile)
+    building.save_intermediate(infile + '.map', 'base_wasm.map')
+    args += ['--output-source-map-url=' + shared.Settings.SOURCE_MAP_BASE + os.path.basename(outfile) + '.map']
     modify_wasm = True
   # tell binaryen to look at the features section, and if there isn't one, to use MVP
   # (which matches what llvm+lld has given us)
-  if shared.Settings.DEBUG_LEVEL >= 2 or shared.Settings.PROFILING_FUNCS or shared.Settings.EMIT_SYMBOL_MAP or shared.Settings.ASYNCIFY_ONLY or shared.Settings.ASYNCIFY_REMOVE or shared.Settings.ASYNCIFY_ADD:
+  if shared.Settings.DEBUG_LEVEL >= 2 or shared.Settings.ASYNCIFY_ADD or shared.Settings.ASYNCIFY_ADVISE or shared.Settings.ASYNCIFY_ONLY or shared.Settings.ASYNCIFY_REMOVE or shared.Settings.EMIT_SYMBOL_MAP or shared.Settings.PROFILING_FUNCS:
     args.append('-g')
   if shared.Settings.WASM_BIGINT:
     args.append('--bigint')
@@ -539,56 +450,42 @@ def finalize_wasm(temp_files, infile, outfile, memfile, DEBUG):
       args.append('--dyncalls-i64')
       # we need to add some dyncalls to the wasm
       modify_wasm = True
-  if not shared.Settings.LEGALIZE_JS_FFI:
-    args.append('--no-legalize-javascript-ffi')
-  else:
+  if shared.Settings.LEGALIZE_JS_FFI:
+    # When we dynamically link our JS loader adds functions from wasm modules to
+    # the table. It must add the original versions of them, not legalized ones,
+    # so that indirect calls have the right type, so export those.
+    if shared.Settings.RELOCATABLE:
+      args.append('--pass-arg=legalize-js-interface-export-originals')
     modify_wasm = True
-  if not shared.Settings.MEM_INIT_IN_WASM:
-    args.append('--separate-data-segments=' + memfile)
+  else:
+    args.append('--no-legalize-javascript-ffi')
+  if memfile:
+    args.append(f'--separate-data-segments={memfile}')
+    args.append(f'--global-base={shared.Settings.GLOBAL_BASE}')
     modify_wasm = True
   if shared.Settings.SIDE_MODULE:
     args.append('--side-module')
-    modify_wasm = True
-  else:
-    # --global-base is used by wasm-emscripten-finalize to calculate the size
-    # of the static data used.  The argument we supply here needs to match the
-    # global based used by lld (see building.link_lld).  For relocatable this is
-    # zero for the global base although at runtime __memory_base is used.
-    # For non-relocatable output we used shared.Settings.GLOBAL_BASE.
-    # TODO(sbc): Can we remove this argument infer this from the segment
-    # initializer?
-    if shared.Settings.RELOCATABLE:
-      args.append('--global-base=0')
-    else:
-      args.append('--global-base=%s' % shared.Settings.GLOBAL_BASE)
   if shared.Settings.STACK_OVERFLOW_CHECK >= 2:
     args.append('--check-stack-overflow')
     modify_wasm = True
   if shared.Settings.STANDALONE_WASM:
     args.append('--standalone-wasm')
-    modify_wasm = True
-  # When we dynamically link our JS loader adds functions from wasm modules to
-  # the table. It must add the original versions of them, not legalized ones,
-  # so that indirect calls have the right type, so export those.
-  if shared.Settings.RELOCATABLE:
-    args.append('--pass-arg=legalize-js-interface-export-originals')
-    modify_wasm = True
+
   if shared.Settings.DEBUG_LEVEL >= 3:
     args.append('--dwarf')
-  stdout = building.run_binaryen_command(
-      'wasm-emscripten-finalize',
-      infile=base_wasm,
-      outfile=wasm if modify_wasm else None,
-      args=args,
-      stdout=subprocess.PIPE)
+  stdout = building.run_binaryen_command('wasm-emscripten-finalize',
+                                         infile=infile,
+                                         outfile=outfile if modify_wasm else None,
+                                         args=args,
+                                         stdout=subprocess.PIPE)
   if modify_wasm:
-    building.save_intermediate(wasm, 'post_finalize.wasm')
-  else:
-    shutil.copyfile(base_wasm, wasm)
-  if write_source_map:
-    building.save_intermediate(wasm + '.map', 'post_finalize.map')
+    building.save_intermediate(infile, 'post_finalize.wasm')
+  elif infile != outfile:
+    shutil.copy(infile, outfile)
+  if shared.Settings.GENERATE_SOURCE_MAP:
+    building.save_intermediate(infile + '.map', 'post_finalize.map')
 
-  if not shared.Settings.MEM_INIT_IN_WASM:
+  if memfile:
     # we have a separate .mem file. binaryen did not strip any trailing zeros,
     # because it's an ABI question as to whether it is valid to do so or not.
     # we can do so here, since we make sure to zero out that memory (even in
@@ -598,7 +495,7 @@ def finalize_wasm(temp_files, infile, outfile, memfile, DEBUG):
   return load_metadata_wasm(stdout, DEBUG)
 
 
-def create_asm_consts_wasm(forwarded_json, metadata):
+def create_asm_consts(metadata):
   asm_consts = {}
   for k, v in metadata['asmConsts'].items():
     const, sigs, call_types = v
@@ -639,27 +536,25 @@ def create_em_js(forwarded_json, metadata):
 
 
 def add_standard_wasm_imports(send_items_map):
-  # Normally we import these into the wasm (so that JS could use them even
-  # before the wasm loads), while in standalone mode we do not depend
-  # on JS to create them, but create them in the wasm and export them.
-  if not shared.Settings.STANDALONE_WASM:
+  if shared.Settings.IMPORTED_MEMORY:
     memory_import = 'wasmMemory'
     if shared.Settings.MODULARIZE and shared.Settings.USE_PTHREADS:
       # Pthreads assign wasmMemory in their worker startup. In MODULARIZE mode, they cannot assign inside the
       # Module scope, so lookup via Module as well.
       memory_import += " || Module['wasmMemory']"
     send_items_map['memory'] = memory_import
-    send_items_map['__indirect_function_table'] = 'wasmTable'
 
   # With the wasm backend __memory_base and __table_base are only needed for
   # relocatable output.
   if shared.Settings.RELOCATABLE:
-    send_items_map['__memory_base'] = str(shared.Settings.GLOBAL_BASE) # tell the memory segments where to place themselves
+    # tell the memory segments where to place themselves
+    send_items_map['__memory_base'] = str(shared.Settings.GLOBAL_BASE)
+    send_items_map['__indirect_function_table'] = 'wasmTable'
+
     # the wasm backend reserves slot 0 for the NULL function pointer
-    table_base = '1'
-    send_items_map['__table_base'] = table_base
+    send_items_map['__table_base'] = '1'
   if shared.Settings.RELOCATABLE:
-    send_items_map['__stack_pointer'] = 'STACK_BASE'
+    send_items_map['__stack_pointer'] = "__stack_pointer"
 
   if shared.Settings.MAYBE_WASM2JS or shared.Settings.AUTODEBUG or shared.Settings.LINKABLE:
     # legalization of i64 support code may require these in some modes
@@ -764,7 +659,7 @@ def add_standard_wasm_imports(send_items_map):
     }'''
 
 
-def create_sending_wasm(invoke_funcs, forwarded_json, metadata):
+def create_sending(invoke_funcs, metadata):
   basic_funcs = []
   if shared.Settings.SAFE_HEAP:
     basic_funcs += ['segfault', 'alignfault']
@@ -774,8 +669,6 @@ def create_sending_wasm(invoke_funcs, forwarded_json, metadata):
   send_items = set(basic_funcs + invoke_funcs + em_js_funcs + declared_items)
 
   def fix_import_name(g):
-    if g.startswith('Math_'):
-      return g.split('_')[1]
     # Unlike fastcomp the wasm backend doesn't use the '_' prefix for native
     # symbols.  Emscripten currently expects symbols to start with '_' so we
     # artificially add them to the output of emscripten-wasm-finalize and them
@@ -796,14 +689,17 @@ def create_sending_wasm(invoke_funcs, forwarded_json, metadata):
   add_standard_wasm_imports(send_items_map)
 
   sorted_keys = sorted(send_items_map.keys())
-  return '{ ' + ', '.join('"' + k + '": ' + send_items_map[k] for k in sorted_keys) + ' }'
+  return '{\n  ' + ',\n  '.join('"' + k + '": ' + send_items_map[k] for k in sorted_keys) + '\n}'
 
 
 def make_export_wrappers(exports, delay_assignment):
   wrappers = []
   for name in exports:
     mangled = asmjs_mangle(name)
-    if shared.Settings.ASSERTIONS:
+    # The emscripten stack functions are called very early (by writeStackCookie) before
+    # the runtime is initialized so we can't create these wrappers that check for
+    # runtimeInitialized.
+    if shared.Settings.ASSERTIONS and not name.startswith('emscripten_stack_'):
       # With assertions enabled we create a wrapper that are calls get routed through, for
       # the lifetime of the program.
       if delay_assignment:
@@ -833,7 +729,7 @@ var %(mangled)s = Module["%(mangled)s"] = asm["%(name)s"]
   return wrappers
 
 
-def create_receiving_wasm(exports, initializers):
+def create_receiving(exports, initializers):
   # When not declaring asm exports this section is empty and we instead programatically export
   # symbols on the global object by calling exportAsmFunctions after initialization
   if not shared.Settings.DECLARE_ASM_MODULE_EXPORTS:
@@ -873,13 +769,15 @@ def create_receiving_wasm(exports, initializers):
   else:
     receiving += make_export_wrappers(exports, delay_assignment)
 
-  return '\n'.join(receiving) + '\n'
+  if shared.Settings.MINIMAL_RUNTIME:
+    return '\n  '.join(receiving) + '\n'
+  else:
+    return '\n'.join(receiving) + '\n'
 
 
-def create_module_wasm(sending, receiving, invoke_funcs, metadata):
+def create_module(sending, receiving, invoke_funcs, metadata):
   invoke_wrappers = create_invoke_wrappers(invoke_funcs)
   receiving += create_named_globals(metadata)
-  receiving += create_fp_accessors(metadata)
   module = []
 
   module.append('var asmLibraryArg = %s;\n' % (sending))
@@ -902,7 +800,6 @@ def load_metadata_wasm(metadata_raw, DEBUG):
     raise
 
   metadata = {
-    'aliases': {},
     'declares': [],
     'implementedFunctions': [],
     'externs': [],
@@ -981,20 +878,17 @@ def generate_struct_info():
   def generate_struct_info():
     with ToolchainProfiler.profile_block('gen_struct_info'):
       out = shared.Cache.get_path(generated_struct_info_name)
-      gen_struct_info.main(['-q', '-c', '-o', out])
+      gen_struct_info.main(['-q', '-o', out])
       return out
 
   shared.Settings.STRUCT_INFO = shared.Cache.get(generated_struct_info_name, generate_struct_info)
 
 
-def run(infile, outfile, memfile):
+def run(in_wasm, out_wasm, outfile_js, memfile):
   temp_files = shared.configuration.get_temp_files()
-  infile, outfile = substitute_response_files([infile, outfile])
   if not shared.Settings.BOOTSTRAPPING_STRUCT_INFO:
     generate_struct_info()
 
-  outfile_obj = open(outfile, 'w')
-
   return temp_files.run_and_clean(lambda: emscript(
-      infile, outfile_obj, memfile, temp_files, shared.DEBUG)
+      in_wasm, out_wasm, outfile_js, memfile, temp_files, shared.DEBUG)
   )
