@@ -214,6 +214,12 @@ var MEMORY_GROWTH_GEOMETRIC_CAP = 96*1024*1024;
 // and MEMORY_GROWTH_GEOMETRIC_CAP are ignored.
 var MEMORY_GROWTH_LINEAR_STEP = -1;
 
+// The "architecture" to compile for. 0 means the default wasm32, 1 is
+// the full end-to-end wasm64 mode, and 2 is wasm64 for clang/lld but lowered to
+// wasm32 in Binaryen (such that it can run on wasm32 engines, while internally
+// using i64 pointers).
+var MEMORY64 = 0;
+
 // If true, allows more functions to be added to the table at runtime. This is
 // necessary for dynamic linking, and set automatically in that mode.
 var ALLOW_TABLE_GROWTH = 0;
@@ -298,6 +304,9 @@ var SYSCALL_DEBUG = 0;
 // Log out socket/network data transfer.
 var SOCKET_DEBUG = 0;
 
+// Log dynamic linker information
+var DYLINK_DEBUG = 0;
+
 // Select socket backend, either webrtc or websockets. XXX webrtc is not
 // currently tested, may be broken
 
@@ -585,6 +594,7 @@ var NODEJS_CATCH_EXIT = 1;
 // then exit with 0 (which hides the error if you don't read the log). With
 // this, we catch any unhandled rejection and throw an actual error, which will
 // make the process exit immediately with a non-0 return code.
+// This should be fixed in Node 15+.
 var NODEJS_CATCH_REJECTION = 1;
 
 // Whether to transform the code using asyncify. This makes it possible to
@@ -950,6 +960,10 @@ var DETERMINISTIC = 0;
 // resolved with the module instance when it is safe to run the compiled code,
 // similar to the `onRuntimeInitialized` callback. You do not need to use the
 // `onRuntimeInitialized` callback when using `MODULARIZE`.
+//
+// (If WASM_ASYNC_COMPILATION is off, that is, if compilation is
+// *synchronous*, then it would not make sense to return a Promise, and instead
+// the Module object itself is returned, which is ready to be used.)
 // 
 // The default name of the function is `Module`, but can be changed using the
 // `EXPORT_NAME` option. We recommend renaming it to a more typical name for a
@@ -1097,9 +1111,8 @@ var WASM = 1;
 // means we don't need to legalize for JS (but this requires a new enough JS
 // VM).
 //
-// Standlone builds by default require a `main` entry point.  If you want to
-// build a library (also known as a reactor) instead you can pass `--no-entry`
-// or specify a list of EXPORTED_FUNCTIONS that does not include `main`.
+// Standlone builds require a `main` entry point by default.  If you want to
+// build a library (also known as a reactor) instead you can pass `--no-entry`.
 var STANDALONE_WASM = 0;
 
 // Whether to ignore implicit traps when optimizing in binaryen.  Implicit
@@ -1233,7 +1246,8 @@ var SDL2_MIXER_FORMATS = ["ogg"];
 var IN_TEST_HARNESS = 0;
 
 // If true, enables support for pthreads.
-// [compile+link] - affects user code at compile and system libraries at link
+// [compile+link] - affects user code at compile and system libraries at link.
+// This setting is equivalent to `-pthread`, which should be preferred.
 var USE_PTHREADS = 0;
 
 // In web browsers, Workers cannot be created while the main browser thread
@@ -1496,6 +1510,9 @@ var RUNTIME_FUNCS_TO_IMPORT = ['abort', 'setTempRet0', 'getTempRet0']
 // not available.  If you are using C++ exceptions, but do not need
 // setjmp()+longjmp() API, then you can set this to 0 to save a little bit of
 // code size and performance when catching exceptions.
+// [compile+link] - at compile time this enables the transformations needed for
+// longjmp support at codegen time, while at link it allows linking in the
+// library support.
 var SUPPORT_LONGJMP = 1;
 
 // If set to 1, disables old deprecated HTML5 API event target lookup behavior.
@@ -1530,9 +1547,10 @@ var MINIFY_HTML = 1;
 // bisecting.
 var MAYBE_WASM2JS = 0;
 
-// The size of our shadow memory.
-// By default, we have 32 MiB. This supports 256 MiB of real memory.
-var ASAN_SHADOW_SIZE = 33554432;
+// This option is no longer used. The appropriate shadow memory size is now
+// calculated from INITIAL_MEMORY and MAXIMUM_MEMORY. Will be removed in a
+// future release.
+var ASAN_SHADOW_SIZE = -1
 
 // Internal: Tracks whether Emscripten should link in exception throwing (C++
 // 'throw') support library. This does not need to be set directly, but pass
@@ -1610,6 +1628,34 @@ var ERROR_ON_WASM_CHANGES_AFTER_LINK = 0;
 // cost of a few bytes extra.
 var ABORT_ON_WASM_EXCEPTIONS = 0;
 
+// Build binaries that use as many WASI APIs as possible, and include additional
+// JS support libraries for those APIs.  This allows emscripten to produce binaries
+// are more WASI compilant and also allows it to process and execute WASI
+// binaries built with other SDKs (e.g.  wasi-sdk).
+// This setting is experimental and subject to change or removal.
+// Implies STANDALONE_WASM.
+var PURE_WASI = 0;
+
+// Set to 1 to define the WebAssembly.Memory object outside of the wasm
+// module.  By default the wasm module defines the memory and exports
+// it to JavaScript.
+// Use of the following settings will enable this settings since they
+// depend on being able to define the memory in JavaScript:
+// - USE_PTHREADS
+// - RELOCATABLE
+// - ASYNCIFY_LAZY_LOAD_CODE
+// - WASM2JS (WASM=0)
+var IMPORTED_MEMORY = 0;
+
+// Generate code to loading split wasm modules.
+// This option will automatically generate two wasm files as output, one
+// with the `.orig` suffix and one without.  The default file (without
+// the suffix) when run will generate instrumentation data can later be
+// fed into wasm-split (the binaryen tool).
+// As well as this the generated JS code will contains help functions
+// to loading split modules.
+var SPLIT_MODULE = 0;
+
 //===========================================
 // Internal, used for testing only, from here
 //===========================================
@@ -1620,15 +1666,6 @@ var OFFSCREEN_FRAMEBUFFER_FORBID_VAO_PATH = 0;
 // Internal (testing only): Forces memory growing to fail.
 var TEST_MEMORY_GROWTH_FAILS = 0;
 
-// Advanced: Customize this array to reduce the set of asm.js runtime variables
-// that are generated. This allows minifying extra bit of asm.js code from unused
-// runtime code, if you know some of these are not needed.
-// (think of this as advanced manual DCE)
-// TODO(sbc): Move to settings_internal (current blocked due to use in test
-// code).
-var ASM_PRIMITIVE_VARS = ['__THREW__', 'threwValue', 'setjmpId', 'tempInt', 'tempBigInt', 'tempBigIntS', 'tempValue', 'tempDouble', 'tempFloat', 'tempDoublePtr', 'STACKTOP', 'STACK_MAX']
-
-// Legacy settings that have been removed or renamed.
 // For renamed settings the format is:
 // [OLD_NAME, NEW_NAME]
 // For removed settings (which now effectively have a fixed value and can no
@@ -1693,4 +1730,5 @@ var LEGACY_SETTINGS = [
   ['EXPORT_FUNCTION_TABLES', [0], 'No longer needed'],
   ['BINARYEN_SCRIPTS', [""], 'No longer needed'],
   ['WARN_UNALIGNED', [0, 1], 'No longer needed'],
+  ['ASM_PRIMITIVE_VARS', [[]], 'No longer needed'],
 ];
