@@ -4,8 +4,7 @@
 # University of Illinois/NCSA Open Source License.  Both these licenses can be
 # found in the LICENSE file.
 
-'''
-A tool that generates FS API calls to generate a filesystem, and packages the files
+"""A tool that generates FS API calls to generate a filesystem, and packages the files
 to work with that.
 
 This is called by emcc. You can also call it yourself.
@@ -22,7 +21,7 @@ data downloads.
 
 Usage:
 
-  file_packager.py TARGET [--preload A [B..]] [--embed C [D..]] [--exclude E [F..]]] [--js-output=OUTPUT.js] [--no-force] [--use-preload-cache] [--indexedDB-name=EM_PRELOAD_CACHE] [--no-heap-copy] [--separate-metadata] [--lz4] [--use-preload-plugins]
+  file_packager TARGET [--preload A [B..]] [--embed C [D..]] [--exclude E [F..]]] [--js-output=OUTPUT.js] [--no-force] [--use-preload-cache] [--indexedDB-name=EM_PRELOAD_CACHE] [--separate-metadata] [--lz4] [--use-preload-plugins]
 
   --preload  ,
   --embed    See emcc --help for more details on those options.
@@ -30,7 +29,7 @@ Usage:
   --exclude E [F..] Specifies filename pattern matches to use for excluding given files from being added to the package.
                     See https://docs.python.org/2/library/fnmatch.html for syntax.
 
-  --from-emcc Indicate that `file_packager.py` was called from `emcc.py` and will be further processed by it, so some code generation can be skipped here
+  --from-emcc Indicate that `file_packager` was called from `emcc` and will be further processed by it, so some code generation can be skipped here
 
   --js-output=FILE Writes output in FILE, if not specified, standard output is used.
 
@@ -42,10 +41,6 @@ Usage:
 
   --indexedDB-name Use specified IndexedDB database name (Default: 'EM_PRELOAD_CACHE')
 
-  --no-heap-copy If specified, the preloaded filesystem is not copied inside the Emscripten HEAP, but kept in a separate typed array outside it.
-                 The default, if this is not specified, is to embed the VFS inside the HEAP, so that mmap()ing files in it is a no-op.
-                 Passing this flag optimizes for fread() usage, omitting it optimizes for mmap() usage.
-
   --separate-metadata Stores package metadata separately. Only applicable when preloading and js-output file is specified.
 
   --lz4 Uses LZ4. This compresses the data using LZ4 when this utility is run, then the client decompresses chunks on the fly, avoiding storing
@@ -58,9 +53,8 @@ Notes:
 
   * The file packager generates unix-style file paths. So if you are on windows and a file is accessed at
     subdir\file, in JS it will be subdir/file. For simplicity we treat the web platform as a *NIX.
-'''
+"""
 
-from __future__ import print_function
 import os
 import sys
 import shutil
@@ -81,7 +75,7 @@ import fnmatch
 import json
 
 if len(sys.argv) == 1:
-  print('''Usage: file_packager.py TARGET [--preload A [B..]] [--embed C [D..]] [--exclude E [F..]]] [--js-output=OUTPUT.js] [--no-force] [--use-preload-cache] [--indexedDB-name=EM_PRELOAD_CACHE] [--no-heap-copy] [--separate-metadata] [--lz4] [--use-preload-plugins]
+  print('''Usage: file_packager TARGET [--preload A [B..]] [--embed C [D..]] [--exclude E [F..]]] [--js-output=OUTPUT.js] [--no-force] [--use-preload-cache] [--indexedDB-name=EM_PRELOAD_CACHE] [--separate-metadata] [--lz4] [--use-preload-plugins]
 See the source for more details.''')
   sys.exit(0)
 
@@ -177,12 +171,6 @@ def main():
   # offline cache instead.
   use_preload_cache = False
   indexeddb_name = 'EM_PRELOAD_CACHE'
-  # If set to True, the blob received from XHR is moved to the Emscripten HEAP,
-  # optimizing for mmap() performance (if ALLOW_MEMORY_GROWTH=0).
-  # If set to False, the XHR blob is kept intact, and fread()s etc. are performed
-  # directly to that data. This optimizes for minimal memory usage and fread()
-  # performance.
-  heap_copy = True
   # If set to True, the package metadata is stored separately from js-output
   # file which makes js-output file immutable to the package content changes.
   # If set to False, the package metadata is stored inside the js-output file
@@ -209,7 +197,7 @@ def main():
       indexeddb_name = arg.split('=', 1)[1] if '=' in arg else None
       leading = ''
     elif arg == '--no-heap-copy':
-      heap_copy = False
+      print('ignoring legacy flag --no-heap-copy (that is the only mode supported now)')
       leading = ''
     elif arg == '--separate-metadata':
       separate_metadata = True
@@ -274,8 +262,13 @@ def main():
           'so that it includes support for loading this file package',
           file=sys.stderr)
 
+  if jsoutput and os.path.abspath(jsoutput) == os.path.abspath(data_target):
+    print('error: TARGET should not be the same value of --js-output',
+          file=sys.stderr)
+    return 1
+
   ret = ''
-  # emcc.py will add this to the output itself, so it is only needed for
+  # emcc will add this to the output itself, so it is only needed for
   # standalone calls
   if not from_emcc:
     ret = '''
@@ -497,16 +490,7 @@ def main():
   if has_preloaded:
     if not lz4:
       # Get the big archive and split it up
-      if heap_copy:
-        use_data = '''
-          // copy the entire loaded file into a spot in the heap. Files will refer to slices in that. They cannot be freed though
-          // (we may be allocating before malloc is ready, during startup).
-          var ptr = Module['getMemory'](byteArray.length);
-          Module['HEAPU8'].set(byteArray, ptr);
-          DataRequest.prototype.byteArray = Module['HEAPU8'].subarray(ptr, ptr+byteArray.length);
-    '''
-      else:
-        use_data = '''
+      use_data = '''
           // Reuse the bytearray from the XHR as the source for file reads.
           DataRequest.prototype.byteArray = byteArray;
     '''
